# -*- coding: utf-8 -*-
from Products.SiteAccess.VirtualHostMonster import VirtualHostMonster
from ZPublisher import BeforeTraverse
from ZPublisher.pubevents import PubStart
from base64 import b64encode
from plone.app.layout.navigation.interfaces import INavigationRoot
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.rest.service import Service
from plone.rest.testing import PLONE_REST_INTEGRATION_TESTING
from zope.event import notify
from zope.interface import alsoProvides
from zope.publisher.interfaces.browser import IBrowserView

import unittest


class TestTraversal(unittest.TestCase):

    layer = PLONE_REST_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

    def traverse(self, path="/plone", accept="application/json", method="GET"):
        request = self.layer["request"]
        request.environ["PATH_INFO"] = path
        request.environ["PATH_TRANSLATED"] = path
        request.environ["HTTP_ACCEPT"] = accept
        request.environ["REQUEST_METHOD"] = method
        auth = "%s:%s" % (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        b64auth = b64encode(auth.encode("utf8"))
        request._auth = "Basic %s" % b64auth.decode("utf8")
        notify(PubStart(request))
        return request.traverse(path)

    def test_json_request_on_portal_root_returns_service(self):
        obj = self.traverse()
        self.assertTrue(isinstance(obj, Service), "Not a service")

    def test_json_request_on_portal_root_with_layout_returns_service(self):
        self.portal.setLayout("summary_view")
        obj = self.traverse()
        self.assertTrue(isinstance(obj, Service), "Not a service")

    def test_json_request_on_portal_root_with_default_page_returns_service(self):
        self.portal.invokeFactory("Document", id="doc1")
        self.portal.setDefaultPage("doc1")
        obj = self.traverse()
        self.assertTrue(isinstance(obj, Service), "Not a service")

    def test_json_request_on_content_object_returns_service(self):
        self.portal.invokeFactory("Document", id="doc1")
        obj = self.traverse(path="/plone/doc1")
        self.assertTrue(isinstance(obj, Service), "Not a service")

    def test_html_request_on_portal_root_returns_default_view(self):
        obj = self.traverse(accept="text/html")
        self.assertEqual(self.portal.getDefaultLayout(), obj.__name__)

    def test_html_request_on_portal_root_returns_dynamic_view(self):
        self.portal.setLayout("summary_view")
        obj = self.traverse(accept="text/html")
        self.assertEqual("summary_view", obj.__name__)

    def test_html_request_on_portal_root_returns_default_page(self):
        self.portal.invokeFactory("Document", id="doc1")
        self.portal.setDefaultPage("doc1")
        obj = self.traverse(accept="text/html")
        self.assertEqual("document_view", obj.__name__)

    def test_json_request_on_object_with_multihook(self):
        doc1 = self.portal[self.portal.invokeFactory("Document", id="doc1")]

        # Register a function to be called before traversal
        def btr_test(container, request):
            request._btr_test_called = 1

        doc1.btr_test = btr_test
        nc = BeforeTraverse.NameCaller("btr_test")
        BeforeTraverse.registerBeforeTraverse(doc1, nc, "Document/btr_test")

        obj = self.traverse(path="/plone/doc1")
        self.assertTrue(isinstance(obj, Service), "Not a service")
        self.assertEqual(1, self.request._btr_test_called)

    def test_json_request_on_existing_view_returns_named_service(self):
        obj = self.traverse("/plone/search")
        self.assertTrue(isinstance(obj, Service), "Not a service")

        folder = self.portal[self.portal.invokeFactory("Folder", id="folder1")]
        alsoProvides(folder, INavigationRoot)
        obj = self.traverse("/plone/folder1/search")
        self.assertTrue(isinstance(obj, Service), "Not a service")

    def test_html_request_on_existing_view_returns_view(self):
        obj = self.traverse(path="/plone/search", accept="text/html")
        self.assertFalse(isinstance(obj, Service), "Got a service")

        folder = self.portal[self.portal.invokeFactory("Folder", id="folder1")]
        alsoProvides(folder, INavigationRoot)
        obj = self.traverse(path="/plone/folder1/search", accept="text/html")
        self.assertFalse(isinstance(obj, Service), "Got a service")

    def test_virtual_hosting(self):
        app = self.layer["app"]
        vhm = VirtualHostMonster()
        vhm.id = "virtual_hosting"
        vhm.addToContainer(app)
        obj = self.traverse(
            path="/VirtualHostBase/http/localhost:8080/plone/VirtualHostRoot/"
        )  # noqa
        self.assertTrue(isinstance(obj, Service), "Not a service")
        del app["virtual_hosting"]

    def test_json_request_to_regular_view_returns_view(self):
        obj = self.traverse("/plone/folder_contents")
        self.assertTrue(IBrowserView.providedBy(obj), "IBrowserView expected")

        self.portal[self.portal.invokeFactory("Folder", id="folder1")]
        obj = self.traverse("/plone/folder1/folder_contents")
        self.assertTrue(IBrowserView.providedBy(obj), "IBrowserView expected")

    def test_json_request_to_view_namespace_returns_view(self):
        obj = self.traverse("/plone/@@folder_contents")
        self.assertTrue(IBrowserView.providedBy(obj), "IBrowserView expected")

        self.portal[self.portal.invokeFactory("Folder", id="folder1")]
        obj = self.traverse("/plone/folder1/@@folder_contents")
        self.assertTrue(IBrowserView.providedBy(obj), "IBrowserView expected")



# -*- coding: utf-8 -*-
from Products.CMFCore.interfaces import ISiteRoot
from Products.SiteAccess.VirtualHostMonster import VirtualHostMonster
from ZPublisher.BaseRequest import DefaultPublishTraverse
from plone.rest.interfaces import IAPIRequest
from plone.rest.interfaces import IService
from plone.rest.events import mark_as_api_request
from zope.component import adapter
from zope.component import queryMultiAdapter
from zope.interface import implementer
from zope.publisher.interfaces.browser import IBrowserPublisher
from zope.traversing.interfaces import ITraversable
from Products.CMFCore.interfaces import IContentish


@adapter(ISiteRoot, IAPIRequest)
class RESTTraverse(DefaultPublishTraverse):
    def publishTraverse(self, request, name):
        try:
            obj = super(RESTTraverse, self).publishTraverse(request, name)
            if not IContentish.providedBy(obj) and not IService.providedBy(obj):
                if isinstance(obj, VirtualHostMonster):
                    return obj
                else:
                    raise KeyError
        except KeyError:
            # No object, maybe a named rest service
            service = queryMultiAdapter(
                (self.context, request), name=request._rest_service_id + name
            )
            if service is None:
                # No service, fallback to regular view
                view = queryMultiAdapter((self.context, request), name=name)
                if view is not None:
                    return view
                raise
            return service

        if name.startswith(request._rest_service_id):
            return obj

        # Do not handle view namespace
        if "@@" in request["PATH_INFO"] or "++view++" in request["PATH_INFO"]:
            return obj

        # Wrap object to ensure we handle further traversal
        return RESTWrapper(obj, request)

    def browserDefault(self, request):
        # Called when we have reached the end of the path
        # In our case this means an unamed service
        return self.context, (request._rest_service_id,)


@implementer(ITraversable)
class MarkAsRESTTraverser(object):
    """
    Traversal adapter for the ``++api++`` namespace.
    It marks the request as API request.
    """

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def traverse(self, name_ignored, subpath_ignored):
        mark_as_api_request(self.request, "application/json")
        return self.context


@implementer(IBrowserPublisher)
class RESTWrapper(object):
    """A wrapper for objects traversed during a REST request."""

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self._bpth_called = False

    def __getattr__(self, name):
        # Delegate attribute access to the wrapped object
        # Needed for security checks in ZPublisher traversal
        return getattr(self.context, name)

    def __getitem__(self, name):
        # Delegate key access to the wrapped object
        return self.context[name]

    # MultiHook requries this to be a class attribute
    def __before_publishing_traverse__(self, arg1, arg2=None):
        bpth = getattr(self.context, "__before_publishing_traverse__", False)
        if bpth:
            if not self._bpth_called:
                self._bpth_called = True
                bpth(arg1, arg2)

    def publishTraverse(self, request, name):
        # Try to get an object using default traversal
        adapter = DefaultPublishTraverse(self.context, request)
        try:
            obj = adapter.publishTraverse(request, name)
            if not IContentish.providedBy(obj) and not IService.providedBy(obj):
                raise KeyError

        # If there's no object with the given name, we get a KeyError.
        # In a non-folderish context a key lookup results in an AttributeError.
        except (KeyError, AttributeError):
            # No object, maybe a named rest service
            service = queryMultiAdapter(
                (self.context, request), name=request._rest_service_id + name
            )
            if service is None:
                # No service, fallback to regular view
                view = queryMultiAdapter((self.context, request), name=name)
                if view is not None:
                    return view
                raise
            return service
        else:
            # Wrap object to ensure we handle further traversal
            return RESTWrapper(obj, request)

    def browserDefault(self, request):
        # Called when we have reached the end of the path
        # In our case this means an unamed service
        return self.context, (request._rest_service_id,)
